#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>

#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>

#if defined(__FreeBSD__) || defined(__NetBSD__)
#define C_NEEDS_UNLIMIT
#endif

#ifdef C_NEEDS_UNLIMIT
#include <sys/resource.h>
#endif


#if defined(SysV) || defined(__svr4__)
#include <limits.h>
#include <time.h>
#include <sys/times.h>
#else
#include <sys/resource.h>
#endif

#include "cra_sys.h"

double last_cpustamp = 0.0;
double last_timestamp = 0.0;

int    basetime;

static sigjmp_buf cra_segv_buf;

void io_error(const char *reason);
void io_error(const char *reason)
{
  fprintf(stderr,"Fehler bei %s\n",reason);
  exit(1);
}

void timestamp()
{
  last_timestamp = time_so_far();
  last_cpustamp = cpu_so_far();
}

double cpu_so_far()
{
#if defined(SysV) || defined(__svr4__)
  struct tms tms;

  if (times(&tms) == -1)
    io_error("times");
  return ((double) tms.tms_utime) / ((double) CLK_TCK) +
    ((double) tms.tms_stime) / ((double) CLK_TCK);

#else
  struct rusage rusage;

  getrusage(RUSAGE_SELF, &rusage);
  return
    ((double) rusage.ru_utime.tv_sec) +
      (((double) rusage.ru_utime.tv_usec) / 1000000.0) +
        ((double) rusage.ru_stime.tv_sec) +
          (((double) rusage.ru_stime.tv_usec) / 1000000.0);
#endif
}

double time_so_far()
{
#if defined(SysV) || defined(__svr4__)
  int        val;
  struct tms tms;

  if ((val = times(&tms)) == -1)
    io_error("times");

  return ((double) val) / ((double) CLK_TCK);

#else
  struct timeval tp;

  if (gettimeofday(&tp, (struct timezone *) NULL) == -1)
    io_error("gettimeofday");
  return ((double) (tp.tv_sec - basetime)) +
    (((double) tp.tv_usec) / 1000000.0);
#endif
}

double return_time(double *cpu, double *real)
{
  *cpu = cpu_so_far();
  *real = time_so_far();

  *cpu = *cpu - last_cpustamp;
  *real = *real - last_timestamp;
  return *cpu;
}

void print_time(const char *was, FILE *ts)
{
  FILE *file;

  if (ts==0) file=stderr; else file=ts;

  fprintf(file,"%.2f s (%.1f%%)"
	  ,cpu_so_far()-last_cpustamp
	  ,(cpu_so_far()-last_cpustamp)/(time_so_far()-last_timestamp)*100.0
	  );

  if (was)
    fprintf(file,": `%s'\n",was);
  else
    fprintf(file,"\n");
}

void print_time_stderr()
{
  print_time(0,stderr);
}

void print_time_stdout()
{
  print_time(0,stdout);
}

#if _POSIX_SOURCE || _SVR4_SOURCE || __svr4__
void cra_bsdsignal(int signo, void *alarm_catcher)
{

    struct sigaction act;
    act.sa_handler = alarm_catcher;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_RESTART;
    if(sigaction(signo, &act,NULL) == -1) {
        perror("signal:");
	/*        return(SIG_ERR); */
    }
}
#else /* sysvr4-quatsch */
void cra_bsdsignal(int signo, void *alarm_catcher)
{
signal(signo,alarm_catcher);
}
#endif  /* syvr4-quatsch */

#ifndef _WIN32
void cra_settimer(int for_what_time, long sec, long usec, void (*function)(int))
{
  static struct itimerval timer1;
  static struct itimerval timerdummy;

  timer1.it_interval.tv_sec=0;
  timer1.it_interval.tv_usec=0;

  timer1.it_value.tv_sec=sec;
  timer1.it_value.tv_usec=usec;

  switch (for_what_time) {
  case CRA_SETTIMER_REAL: {
    setitimer(ITIMER_REAL,&timer1,&timerdummy);
    cra_bsdsignal(SIGALRM,function);
    break;
  }
  case CRA_SETTIMER_USER: {
    setitimer(ITIMER_VIRTUAL,&timer1,&timerdummy);
    cra_bsdsignal(SIGVTALRM,function);
    break;
  }
  case CRA_SETTIMER_USERSYS: {
    setitimer(ITIMER_PROF,&timer1,&timerdummy);
    cra_bsdsignal(SIGPROF,function);
    break;
  }
  }
}
#endif

/* Returns a random number 0 <= number < max */
#ifndef RAND_MAX
#define RAND_MAX 2147483647
#endif

long cra_irand ( long max)
{
long long l2 = max;
long long l3 = RAND_MAX;
long long l1 = rand();

return ( l1 * l2 ) / l3;
}
/* floating point version */
double cra_frand ( double max)
{
return ((double)rand() * max) / (double)RAND_MAX;
}

void unlimit()
{
#ifdef C_NEEDS_UNLIMIT
  int i;
  struct rlimit rl;
  struct rlimit rlo;
  int limits[] = {
    RLIMIT_CORE,
    RLIMIT_CPU,
    RLIMIT_DATA,
    RLIMIT_FSIZE,
    RLIMIT_MEMLOCK,
    RLIMIT_NOFILE,
    RLIMIT_NPROC,
    RLIMIT_RSS,
    RLIMIT_STACK
  };
#define NLIMITS 9 

  rl.rlim_cur=RLIM_INFINITY;
  for (i=0 ; i<NLIMITS ; i++) {
    getrlimit(limits[i],&rlo);
    rl.rlim_max = rlo.rlim_max;
    if (getrlimit(limits[i],&rl))
      fprintf(stderr,"Couldn't raise shell limits\n");
  }
#endif
}

/* Helper routines for cra_pifvalid */
static void 
cra_segv_handler(int sig)
{
  siglongjmp(cra_segv_buf,sig);
}

static void
cra_handle_segv(void)
{
  struct sigaction sig;
  
  sig.sa_handler = cra_segv_handler;
  sig.sa_flags = SA_RESTART;
  sigemptyset(&sig.sa_mask);
  if (sigaction(SIGSEGV,&sig,NULL) < 0) {
    perror("sigaction sigsegv");
    exit(2);
  }
  if (sigaction(SIGBUS,&sig,NULL) < 0) {
    perror("sigaction sigbus");
    exit(2);
  }
}

static void 
cra_unhandle_segv(void)
{
  struct sigaction sig;
  
  sig.sa_handler = NULL;
  sig.sa_flags = SA_RESTART;
  sigemptyset(&sig.sa_mask);
  if (sigaction(SIGSEGV,&sig,NULL) < 0) {
    perror("sigaction sigsegv");
    exit(2);
  }
  if (sigaction(SIGBUS,&sig,NULL) < 0) {
    perror("sigaction sigbus");
    exit(2);
  }
}


/* This routine processes string s, if the pointer doesn't cause a 
   segfault or bus error. If it isn't valid, it processes
   the string `if_not_valid' instead.

   If function pointer fun is not NULL, the function will be called
   with the string to be processed. If fun is NULL, the string is just
   fputs()'ed to file f.

   */

int
cra_pifvalid(FILE *f, const char *const s, const char *const if_not_valid
	     ,void (*fun)(const char *const, FILE *f))
{
  volatile int c = 0;
  int which_sig;
  const char *message;
  const char *sp;

  /* First see if its valid */
  cra_handle_segv();
  if ( (which_sig = sigsetjmp(cra_segv_buf,0)) == 0) {
    sp = s;
    while (*sp) {
      c += *sp;
      sp++;
    }
    message = s;
  }
  else {
    cra_unhandle_segv();
    if (if_not_valid)
      message = if_not_valid;
    else
      return c;
  }
  cra_unhandle_segv();
  if (fun)
    fun(message,f);
  else
    fputs(message,f);
  return c; /* Prevent optimization */
}
